using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.SqlClient;
using System.Drawing;
using System.IO;
using System.Text;
using System.Transactions;
using System.Windows.Forms;
using Microsoft.KtmIntegration;

namespace TxFAndSql
{
    public partial class Form1 : Form
    {
        private TransactionScope myTransactionScope;
        private FileInfo myTransactedFile = new FileInfo(Path.Combine(Directory.GetCurrentDirectory(), @"FileStore\testTransactedFile.txt"));
        public static string myConnectionString = string.Format(@"Data Source=.;AttachDbFilename={0}\TestDatabase.mdf;Integrated Security=True;", Directory.GetCurrentDirectory());

        #region Constructors

        public Form1()
        {
            InitializeComponent();

            // Clean up and get files ready
            if (!myTransactedFile.Directory.Exists)
                myTransactedFile.Directory.Create();
            
            if (myTransactedFile.Exists)
                myTransactedFile.Delete();

            // Clean up SQL
            using (SqlConnection myConnection = new SqlConnection(myConnectionString))
            {
                myConnection.Open();
                SqlCommand myCommand = new SqlCommand(
                    "TRUNCATE TABLE tblDocument",
                    myConnection);

                myCommand.ExecuteNonQuery();
            }
        }

        #endregion 

        #region Transaction Management

        private void startButton_Click(object sender, EventArgs e)
        {
            if (Transaction.Current == null)
            {
                myTransactionScope = new TransactionScope();

                statusLabel.Text = "Transaction Started...";
            }
            else
            {
                statusLabel.Text = "Transaction Already Started...";
            }
        }

        private void commitButton_Click(object sender, EventArgs e)
        {
            if (myTransactionScope != null)
            {
                myTransactionScope.Complete();
                myTransactionScope.Dispose();

                statusLabel.Text = "Committed Transaction";
            }
        }

        private void rollbackButton_Click(object sender, EventArgs e)
        {
            // If we are in a transaction, just dispose the scope to rollback
            if (Transaction.Current != null)
            {
                myTransactionScope.Dispose();

                statusLabel.Text = "Rolled Transaction Back...";
            }
        }

        #endregion

        private void transactedFileButton_Click(object sender, EventArgs e)
        {
            if (Transaction.Current == null)
            {
                throw new Exception("Must be within a transaction scope");
            }
    
            using (TransactionScope fileTransaction = new TransactionScope(
                Transaction.Current))
            {
                statusLabel.Text = "Creating Transacted File...";

                using (StreamWriter myFile = new StreamWriter(TransactedFile.Open(
                        myTransactedFile.FullName,
                        FileMode.Create,
                        FileAccess.Write,
                        FileShare.None)))
                {

                    myFile.Write("I am a transacted file");
                }

                fileTransaction.Complete();
                statusLabel.Text = "Created Transacted File...";
            }
        }

        private void sqlButton_Click(object sender, EventArgs e)
        {
            if (Transaction.Current == null)
            {
                throw new Exception("Must be within a transaction scope");
            }

            using (SqlConnection myConnection = new SqlConnection(myConnectionString))
            {
                myConnection.Open();

                // Build SQL Command
                string sql = "INSERT INTO tblDocument (Name, DateCreated) VALUES (@name, @dateCreated)";
                SqlCommand myCommand = new SqlCommand(sql,
                    myConnection);
                
                myCommand.Parameters.Add(
                    new SqlParameter("@name", myTransactedFile.FullName));
                myCommand.Parameters.Add(
                    new SqlParameter("@dateCreated", DateTime.Now));

                myCommand.ExecuteNonQuery();
            }

            statusLabel.Text = "Executed SQL...";
        }

        protected override void OnClosed(EventArgs e)
        {
            if (Transaction.Current != null)
                myTransactionScope.Dispose();

            base.OnClosed(e);
        }
    }
}