using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Transactions;
using System.Windows.Forms;
using Microsoft.KtmIntegration;

namespace TxFOnly
{
    public partial class Form1 : Form
    {
        private FileInfo myTransactedFile = new FileInfo(
            Path.Combine(Directory.GetCurrentDirectory(), @"FileStore\testTransactedFile.txt"));

        private FileInfo myNonTransactedFile = new FileInfo(
            Path.Combine(Directory.GetCurrentDirectory(), @"FileStore\testNonTransactedFile.txt"));

        private TransactionScope myTransactionScope;

        #region Constructors

        public Form1()
        {
            InitializeComponent();

            if (!myTransactedFile.Directory.Exists)
                myTransactedFile.Directory.Create();
            
            if (myTransactedFile.Exists)
                myTransactedFile.Delete();

            if (myNonTransactedFile.Exists)
                myNonTransactedFile.Delete();
        }

        #endregion 

        #region Transaction Management

        private void startButton_Click(object sender, EventArgs e)
        {
            if (Transaction.Current == null)
            {
                myTransactionScope = new TransactionScope();

                statusLabel.Text = "Transaction Started...";
            }
            else
            {
                statusLabel.Text = "Transaction Already Started...";
            }
        }

        private void commitButton_Click(object sender, EventArgs e)
        {
            if (myTransactionScope != null)
            {
                myTransactionScope.Complete();
                myTransactionScope.Dispose();

                statusLabel.Text = "Committed Transaction";
            }
        }

        private void rollbackButton_Click(object sender, EventArgs e)
        {
            // If we are in a transaction, just dispose the scope to rollback
            if (Transaction.Current != null)
            {
                myTransactionScope.Dispose();

                statusLabel.Text = "Rolled Transaction Back...";
            }
        }

        #endregion

        private void transactedFileButton_Click(object sender, EventArgs e)
        {
            if (Transaction.Current == null)
            {
                throw new Exception("Must be within a transaction scope");
            }
    
            using (TransactionScope fileTransaction = 
                new TransactionScope(Transaction.Current))
            {
                statusLabel.Text = "Creating Transacted File...";

                using (StreamWriter myFile = 
                    new StreamWriter(TransactedFile.Open(
                        myTransactedFile.FullName,
                        FileMode.Create,
                        FileAccess.Write,
                        FileShare.None)))
                {

                    myFile.Write("I am a transacted file");
                }

                fileTransaction.Complete();
                statusLabel.Text = "Created Transacted File...";
            }
        }

        private void fileButton_Click(object sender, EventArgs e)
        {
            using (StreamWriter myFile = new StreamWriter(File.Open(myNonTransactedFile.FullName, FileMode.Create)))
            {
                myFile.Write("I am NOT a transacted file");
            }

            statusLabel.Text = "Created NON-Transacted File...";
        }

        protected override void OnClosed(EventArgs e)
        {
            if (Transaction.Current != null)
                myTransactionScope.Dispose();

            base.OnClosed(e);
        }
    }
}